import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.etl.transformers import DataFrameFilterTransformer
from spetlr.spark import Spark


class TestDataFrameFilterTransformer(DataframeTestCase):
    def test_filter_with_column_name_and_value(self):
        """
        This test accepts col_name and col_value in DataFrameFilterTransformer and
        checks if the output dataframe is correctly filtered.
        """
        input_schema = T.StructType(
            [
                T.StructField("Col1", T.StringType(), True),
                T.StructField("Col2", T.IntegerType(), True),
                T.StructField("Col3", T.DoubleType(), True),
                T.StructField("Col4", T.StringType(), True),
                T.StructField("Col5", T.StringType(), True),
            ]
        )

        input_data1 = ("Col1Data", 42, 13.37, "Col4Data", "Col5Data")
        input_data2 = ("Col1Data_2nd", 43, 23.37, "Col4Data_2nd", "Col5Data_2nd")
        input_data3 = ("Col1Data", 45, 20.15, "Col4Data_3rd", "Col5Data_3rd")

        input_data = [input_data1, input_data2, input_data3]

        input_df = Spark.get().createDataFrame(data=input_data, schema=input_schema)

        transformed_df = DataFrameFilterTransformer(
            col_value="Col1Data", col_name="Col1"
        ).process(input_df)

        expected_data = [
            input_data1,
            input_data3,
        ]

        self.assertDataframeMatches(
            df=transformed_df,
            expected_data=expected_data,
        )

    def test_filter_with_default_column_name(self):
        """
        This test accepts only col_value and uses the default col_name='messageType' in
        DataFrameFilterTransformer and checks if the output dataframe is correctly
        filtered.
        """
        input_schema = T.StructType(
            [
                T.StructField("messageType", T.StringType(), True),
                T.StructField("Col2", T.IntegerType(), True),
                T.StructField("Col3", T.DoubleType(), True),
                T.StructField("Col4", T.StringType(), True),
                T.StructField("Col5", T.StringType(), True),
            ]
        )

        input_data1 = ("Col1Data", 42, 13.37, "Col4Data", "Col5Data")
        input_data2 = ("Col1Data_2nd", 42, 13.37, "Col4Data_2nd", "Col5Data_2nd")
        input_data3 = ("Col1Data", 45, 20.15, "Col4Data_3rd", "Col5Data_3rd")

        input_data = [input_data1, input_data2, input_data3]

        input_df = Spark.get().createDataFrame(data=input_data, schema=input_schema)

        transformed_df = DataFrameFilterTransformer(
            col_value="Col1Data",
        ).process(input_df)

        expected_data = [
            input_data1,
            input_data3,
        ]

        self.assertDataframeMatches(
            df=transformed_df,
            expected_data=expected_data,
        )



import unittest

from pyspark.sql import types as t

import spetlr.spark
from spetlr.etl.transformers.fuzzy_select import FuzzySelectTransformer


class FuzzySelectorTest(unittest.TestCase):
    def test_fuzzy1(self):
        target_columns = [
            "unChanged",
            "upperCase",
            "idStringId",
            "idStringABCProductId",
            "misSpelled",
        ]

        in_columns = [
            "unChanged",
            "UpperCase",
            "IDStringID",
            "IDStringABCProductId",  # this column is similar enough to the one above
            # that the standard match cutoff of 0.6 fails to associate it uniquely.
            "miisSpolled",
        ]

        res = dict(zip(in_columns, target_columns))

        self.assertRaises(
            spetlr.etl.transformers.fuzzy_select.NonUniqueException,
            FuzzySelectTransformer(target_columns).find_best_mapping,
            in_columns,
        )

        self.assertEqual(
            FuzzySelectTransformer(
                target_columns,
                # a tighter similarity constraint allows a unique association
                match_cutoff=0.8,
            ).find_best_mapping(
                in_columns,
            ),
            res,
        )

    def test_transform(self):
        ft = FuzzySelectTransformer(
            [
                "Index",
                "Count",
                "Label",
            ]
        )

        result = ft.process(
            spetlr.spark.Spark.get().createDataFrame(
                [(1, 2, "foo"), (3, 4, "bar")],
                t.StructType(
                    [
                        t.StructField("inex", t.IntegerType()),
                        t.StructField("count", t.IntegerType()),
                        t.StructField("lables", t.StringType()),
                    ]
                ),
            )
        )

        self.assertEqual(
            [
                "Index",
                "Count",
                "Label",
            ],
            result.columns,
        )



import warnings
from typing import List

from pyspark.sql import DataFrame

from .types import EtlBase, dataset_group


class Transformer(EtlBase):
    """
    This is the base Transformer class for the OETL framework.

    If you only want to transform a single input dataframe,
    implement `process`
    If you want to transform a set of dataframes,
    implement `process_many`

    In regards to the etl step, the behavior of the transformation is controlled
    by the parameters dataset_input_keys, dataset_output_key and consume_inputs.

    Attributes:
    ----------
        dataset_input_keys : Union[str, List[str]]
            str or list of input dataset keys that will be used for
            either process or process_many
        dataset_output_key : str
            name of the output key that we be added to the dataset after transformation
        consume_inputs : bool
            Flag to control dataset consuming behavior
    """

    def __init__(
        self,
        dataset_key: str = None,
        *,
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ):
        if dataset_key is not None:
            warnings.warn(
                "The parameter dataset_key is deprecated, use dataset_output_key instead."  # noqa: E501
            )
            dataset_output_key = dataset_key

        self.dataset_input_keys = dataset_input_keys

        if dataset_output_key is None:
            self.dataset_output_key = type(self).__name__
        else:
            self.dataset_output_key = dataset_output_key

        self.consume_inputs = consume_inputs

    def etl(self, inputs: dataset_group) -> dataset_group:
        # If dataset_input_keys is None, it will be set to all keys in inputs
        dataset_input_keys = self.dataset_input_keys or list(inputs.keys())

        # Call process or process_many depending on the input_key count
        if len(dataset_input_keys) == 1:
            df = inputs[dataset_input_keys[0]]
            if df is not None:
                df = self.process(df)
        else:
            inputs_filtered = {
                datasetKey: df
                for datasetKey, df in inputs.items()
                if datasetKey in dataset_input_keys
            }
            df = self.process_many(inputs_filtered)

        if self.consume_inputs:
            for key in dataset_input_keys:
                inputs.pop(key)

        # dataset_output_key is added to inputs after pop is called
        # This ensures that dataset_output_key is not popped.
        inputs[self.dataset_output_key] = df
        return inputs

    def process(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError()

    def process_many(self, datasets: dataset_group) -> DataFrame:
        raise NotImplementedError()



## Commenting out as these tests are not working as expected in UC-enabled clusters.


# import unittest

# from spetlrtools.testing import DataframeTestCase

# from spetlr import Configurator
# from spetlr.deltaspec import DeltaTableSpec
# from spetlr.spark import Spark
# from tests.cluster.delta.deltaspec import tables


# @unittest.skipUnless(
#     Spark.version() >= Spark.DATABRICKS_RUNTIME_11_3,
#     "Drop column only supported from DBR 11.0",
# )
# class TestTableSpec(DataframeTestCase):
#     @classmethod
#     def setUpClass(cls) -> None:
#         c = Configurator()
#         c.clear_all_configurations()
#         c.set_debug()

#         c.register("mydb", dict(name="myDeltaTableSpecTestDb{ID}"))
#         cls.base = tables.base
#         cls.target = tables.target

#         db = c.get("mydb", "name")
#         Spark.get().sql(f"CREATE DATABASE {db};")

#     @classmethod
#     def tearDownClass(cls) -> None:
#         c = Configurator()
#         db = c.get("mydb", "name")
#         # clean up after test.
#         Spark.get().sql(f"DROP DATABASE {db} CASCADE")

#     def test_01_tblspec(self):
#         # at first the table does not exist
#         diff = self.base.compare_to_name()
#         self.assertTrue(diff.is_different(), diff)
#         self.assertTrue(diff.nullbase(), diff)

#         # then we make it exist
#         self.base.make_storage_match()

#         # now it exists and matches
#         diff = self.base.compare_to_name()
#         self.assertFalse(diff.is_different(), repr(diff))

#         # but it does not match the target
#         diff = self.target.compare_to_name()
#         # the names are the same
#         self.assertTrue(diff.name_match(), repr(diff))
#         # the rest of the table is not the same
#         self.assertTrue(diff.is_different(), repr(diff))

#         # the table is not readable because of schema mismatch
#         self.assertFalse(self.target.is_readable())

#         # overwriting is possible and updates to the target schema
#         df = Spark.get().createDataFrame([(1, "a", 3.14, "b", "c")], self.target.schema)
#         self.target.make_storage_match(errors_as_warnings=True)

#         self.target.get_dh().overwrite(df, overwriteSchema=True)

#         # now the base no longer matches
#         diff = self.base.compare_to_name()
#         self.assertTrue(diff.is_different(), diff)

#         # but the target matches.
#         diff = self.target.compare_to_name()
#         self.assertFalse(diff.is_different(), repr(diff))

#     def test_02_location_change(self):
#         tables.oldlocation.make_storage_match(allow_table_create=True)

#         # location mismatch makes the tables not readable
#         self.assertFalse(tables.newlocation.is_readable())

#         # overwriting will update the table location.
#         tables.newlocation.make_storage_match(allow_location_change=True)

#         diff = tables.newlocation.compare_to_name()
#         self.assertTrue(diff.complete_match(), diff)
#         self.assertTrue(tables.newlocation.is_readable(), diff)

#     def test_03_name_change(self):
#         tables.oldname.make_storage_match(allow_table_create=True)

#         spark = Spark.get()
#         df = spark.createDataFrame([("eggs", 3.5, "spam")], tables.oldname.schema)
#         tables.oldname.get_dh().overwrite(df)

#         for stmt in tables.newname.compare_to(tables.oldname).alter_statements(
#             allow_name_change=True
#         ):
#             spark.sql(stmt)

#         diff = tables.newname.compare_to_name()
#         self.assertTrue(diff.complete_match(), diff)
#         self.assertEqual(tables.newname.get_dh().read().count(), 1)

#     def test_04_can_control_managed_table(self):
#         tables.managed.make_storage_match(allow_table_create=True)

#         diff = tables.managed.compare_to_name()
#         self.assertTrue(diff.complete_match(), diff)

#     def test_05_modify_to_allow_drop(self):
#         # a table has been created without the DeltaTableSpec:
#         spark = Spark.get()
#         spark.sql(tables.simple_create_sql.format(**Configurator().get_all_details()))

#         # now a modified version is to be applied that drops columns:
#         ds = DeltaTableSpec.from_sql(tables.simple_modified_sql)
#         ds.make_storage_match(allow_columns_drop=True)

#         diff = ds.compare_to_name()
#         self.assertTrue(diff.complete_match(), diff)



"""
This function searches through .sql files and looks for statements to create tables
or databases:
https://spark.apache.org/docs/3.0.0-preview/sql-ref-syntax-ddl-create-database.html
https://spark.apache.org/docs/latest/sql-ref-syntax-ddl-create-table-datasource.html
and returns a dictionary of configuration details.
"""

import importlib.resources
from pathlib import Path
from types import ModuleType
from typing import Dict, Union

from spetlr.configurator.sql.comments import _extract_comment_attributes
from spetlr.configurator.sql.create import _walk_create_statement
from spetlr.configurator.sql.init_sqlparse import parse
from spetlr.exceptions.configurator_exceptions import (
    SpetlrConfiguratorInvalidSqlCommentsException,
)


def parse_sql_code_to_config(sql_code: str) -> Dict:
    details = {}

    # the sequence "-- COMMAND ----------" is used in jupyter notebooks
    # and separates cells.
    # We treat it as another way to end a statement
    sql_code = sql_code.replace("-- COMMAND ----------", ";")

    for statement in parse(sql_code):
        comment_attributes = _extract_comment_attributes(statement)
        if "key" not in comment_attributes:
            # if no magic comments were used on the statement,
            # then there is nothing to configure here
            continue
        table_id = comment_attributes.pop("key")

        object_details = _walk_create_statement(statement)

        for key in object_details:
            if key in comment_attributes:
                raise SpetlrConfiguratorInvalidSqlCommentsException(
                    f"Error for {key} in {table_id}, "
                    "The comments must not specify attributes, "
                    "that can be derived directly from the sql."
                )
        object_details.update(comment_attributes)
        details[table_id] = object_details
    return details


def parse_single_sql_statement(sql_code: str) -> dict:
    """Take a sinle sql statement, such as a create table statement,
    and break it down into its constituent parts, such as name, location and schema,
    which are returned as a dictionary.
    """
    # This will fail unless there is exactly one statement
    (statement,) = parse(sql_code)

    comment_attributes = _extract_comment_attributes(statement)

    object_details = _walk_create_statement(statement)

    comment_attributes.update(object_details)
    return comment_attributes


def _parse_sql_to_config(resource_path: Union[str, ModuleType]) -> Dict:
    details = {}
    for file_name in importlib.resources.contents(resource_path):
        extension = Path(file_name).suffix
        if not extension == ".sql":
            continue
        with importlib.resources.path(resource_path, file_name) as file_path:
            with open(file_path) as file:
                sql_code = file.read()
                details.update(parse_sql_code_to_config(sql_code))
    return details



from pyspark.sql.types import (
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.testing import DataframeTestCase
from spetlrtools.time import dt_utc

from spetlr.etl.transformers.drop_oldest_duplicate_transformer import (
    DropOldestDuplicatesTransformer,
)
from spetlr.spark import Spark


class TestDropOldestDuplicatesTransformer(DataframeTestCase):
    def test_drop_oldest_duplicates_transformer(self):
        data = [
            (1, "Fender", "Telecaster", 5, dt_utc(2021, 7, 1, 10)),
            (1, "Fender", "Telecaster", 4, dt_utc(2021, 7, 1, 11)),
            (2, "Gibson", "Les Paul", 27, dt_utc(2021, 7, 1, 11)),
            (3, "Ibanez", "RG", 22, dt_utc(2021, 8, 1, 11)),
            (3, "Ibanez", "RG", 26, dt_utc(2021, 9, 1, 11)),
            (3, "Ibanez", "RG", 18, dt_utc(2021, 10, 1, 11)),
        ]

        expected_data = [
            [1, "Fender", "Telecaster", 4, dt_utc(2021, 7, 1, 11)],
            [2, "Gibson", "Les Paul", 27, dt_utc(2021, 7, 1, 11)],
            [3, "Ibanez", "RG", 18, dt_utc(2021, 10, 1, 11)],
        ]

        schema = StructType(
            [
                StructField("id", IntegerType(), True),
                StructField("model", StringType(), True),
                StructField("brand", StringType(), True),
                StructField("amount", IntegerType(), True),
                StructField("timecolumn", TimestampType(), True),
            ]
        )

        df = Spark.get().createDataFrame(data=data, schema=schema)

        df2 = (
            DropOldestDuplicatesTransformer(
                cols=["id", "model", "brand"], orderByColumn="timecolumn"
            )
            .process(df)
            .orderBy("id")
        )

        self.assertDataframeMatches(
            df2,
            None,
            expected_data=expected_data,
        )



from spetlrtools.testing import DataframeTestCase

from spetlr import Configurator
from spetlr.delta import DeltaHandle
from spetlr.etl import Orchestrator
from spetlr.etl.extractors import SimpleExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.transformers.SimpleSqlTransformer import SimpleSqlTransformer
from spetlr.spark import Spark
from tests.cluster.transformations import sql


class SimpleSqlTransformerTest(DataframeTestCase):
    @classmethod
    def setUpClass(cls):
        c = Configurator()
        c.clear_all_configurations()
        c.set_debug()
        c.register("db", {"name": "SimpleSqlTransformerTestDb{ID}"})
        c.register("tbl1", {"name": "{db}.tbl1"})
        c.register("tbl2", {"name": "{db}.tbl2"})
        c.register("tbl3", {"name": "{db}.tbl3"})

        db = c.get("db", "name")
        tbl1 = c.get("tbl1", "name")
        tbl2 = c.get("tbl2", "name")
        tbl3 = c.get("tbl3", "name")

        spark = Spark.get()
        spark.sql(f"CREATE DATABASE {db};")
        spark.sql(f"""CREATE TABLE {tbl1} (a int, b int) USING DELTA;""")
        spark.sql(f"""CREATE TABLE {tbl2} (a int, c int) USING DELTA;""")
        spark.sql(f"""CREATE TABLE {tbl3} (a int, b int, c int) USING DELTA;""")

        spark.createDataFrame([(1, 2)], "a int, b int").write.mode(
            "overwrite"
        ).saveAsTable(tbl1)
        spark.createDataFrame([(1, 3)], "a int, c int").write.mode(
            "overwrite"
        ).saveAsTable(tbl2)

    @classmethod
    def tearDownClass(cls) -> None:
        Spark.get().sql(f"""DROP DATABASE {Configurator().get('db','name')} CASCADE""")

    def test_all(self):
        o = Orchestrator()
        o.extract_from(
            SimpleExtractor(DeltaHandle.from_tc("tbl1"), dataset_key="FirstTable")
        )
        o.extract_from(
            SimpleExtractor(DeltaHandle.from_tc("tbl2"), dataset_key="SecondTable")
        )
        o.transform_with(
            SimpleSqlTransformer(
                sql_modue=sql,
                sql_file_pattern="transform",
                dataset_input_keys=["FirstTable", "SecondTable"],
                dataset_output_key="SimpleSqlTransformerTestResult",
            )
        )
        o.load_into(SimpleLoader(DeltaHandle.from_tc("tbl3")))
        o.execute()

        self.assertDataframeMatches(
            DeltaHandle.from_tc("tbl3").read(), None, [(1, 2, 3)]
        )



from typing import List, Union

from pyspark.sql import DataFrame

from .types import EtlBase, dataset_group


class Loader(EtlBase):
    """
    A loader is a data sink, which is indicated by the fact that
    the save method has no return.

    In regards to the etl step, a loader USES the input dataset(s)
    and does not consume or change it.
    """

    def __init__(
        self,
        *,
        dataset_input_keys: Union[str, List[str]] = None,
    ):
        if dataset_input_keys is None:
            self.dataset_input_key_list = []
        elif isinstance(dataset_input_keys, str):
            self.dataset_input_key_list = [dataset_input_keys]
        else:
            self.dataset_input_key_list = dataset_input_keys

    def etl(self, inputs: dataset_group) -> dataset_group:
        if len(self.dataset_input_key_list) > 0:
            if len(self.dataset_input_key_list) == 1:
                df = inputs[self.dataset_input_key_list[0]]
                if df is not None:
                    self.save(df)
            else:
                datasetFilteret = {
                    datasetKey: df
                    for datasetKey, df in inputs.items()
                    if datasetKey in self.dataset_input_key_list
                }
                self.save_many(datasetFilteret)
        elif len(inputs) == 1:
            df = next(iter(inputs.values()))
            if df is not None:
                self.save(df)
        else:
            self.save_many(inputs)

        return inputs

    def save(self, df: DataFrame) -> None:
        raise NotImplementedError()

    def save_many(self, datasets: dataset_group) -> None:
        raise NotImplementedError()



## Commenting out as these tests are not working as expected in UC-enabled clusters.


# import unittest

# from spetlr import Configurator
# from spetlr.configurator.sql.parse_sql import parse_sql_code_to_config
# from spetlr.deltaspec.DeltaDatabaseSpec import DeltaDatabaseSpec
# from spetlr.spark import Spark


# class TestDbSpec(unittest.TestCase):
#     @classmethod
#     def setUpClass(cls):
#         c = Configurator()
#         c.clear_all_configurations()
#         c.set_debug()

#     def test_DbSpecFromDisk(self):
#         c = Configurator()
#         master_sql = """
#             -- SPETLR.CONFIGURATOR key: MyDbAlias
#             CREATE DATABASE IF NOT EXISTS my_db_name{ID}
#             COMMENT "Really great db"
#             LOCATION "/tmp/my_db{ID}";
#             """
#         config = parse_sql_code_to_config(master_sql)

#         k, v = config.popitem()
#         self.assertEqual(k, "MyDbAlias")
#         c.register(k, v)
#         db_parsed = DeltaDatabaseSpec.from_tc(k)
#         Spark.get().sql(master_sql.format(**c.get_all_details()))
#         db_read = DeltaDatabaseSpec.from_spark(db_parsed.fully_substituted().name)

#         self.assertEqual(db_parsed.fully_substituted(), db_read)

#         Spark.get().sql(f"DROP DATABASE {db_parsed.fully_substituted().name} CASCADE")



import unittest
from datetime import datetime

import pandas as pd
from pandas.testing import assert_frame_equal

from spetlr.power_bi.SparkPandasDataFrame import SparkPandasDataFrame


class TestSparkPandasDataFrame(unittest.TestCase):
    def test_get_workspace_success(self):
        # Arrange
        json = [
            {
                "id": 3,
                "refreshType": "ViaEnhancedApi",
                "status": "Completed",
                "startTime": "2024-03-31T00:00:00Z",  # summer time: change from 1
                "endTime": "2024-03-31T02:00:00Z",  # hour to 2 hours difference
                "serviceExceptionJson": None,
                "requestId": "aec28227-f7af-4c2d-a4e6-fcb01cd570ec",
                "refreshAttempts": None,
            },
            {
                "id": 2,
                "refreshType": "ViaApi",
                "status": "Completed",
                "startTime": "2024-02-26T10:00:00Z",
                "endTime": "2024-02-26T10:05:00Z",  # winter time: 1 hour
                "serviceExceptionJson": None,  # difference from UTC
                "requestId": "74d25c0b-0473-4dd9-96ff-3ca737b072a7",
                "refreshAttempts": None,
            },
            {
                "id": 1,
                "refreshType": "OnDemand",
                "status": "Completed",
                "startTime": "2024-01-16T10:00:00Z",
                "endTime": None,  # will calculate None seconds
                "serviceExceptionJson": None,
                "requestId": "191cc73a-fc75-4eff-9ffe-67aa5290d7f4",
                "refreshAttempts": None,
            },
        ]
        schema = [
            ("id", "Id", "long"),
            ("refreshType", "RefreshType", "string"),
            ("status", "Status", "string"),
            (
                lambda d: (d["endTime"] - d["startTime"]) / pd.Timedelta(seconds=1),
                "Seconds",
                "long",
            ),
            ("startTime", "StartTime", "timestamp"),
            ("endTime", "EndTime", "timestamp"),
            ("serviceExceptionJson", "Error", "string"),
            ("requestId", "RequestId", "string"),
            ("refreshAttempts", "RefreshAttempts", "string"),
        ]
        expected = pd.DataFrame(
            {
                "Id": [1, 2, 3],
                "RefreshType": ["OnDemand", "ViaApi", "ViaEnhancedApi"],
                "Status": ["Completed", "Completed", "Completed"],
                "Seconds": [
                    None,
                    300,
                    7200,
                ],  # despite the daylight saving change it
                # correctly shows 7200 seconds !
                "StartTimeLocal": [
                    datetime(2024, 1, 16, 11, 0),
                    datetime(2024, 2, 26, 11, 0),
                    datetime(2024, 3, 31, 1, 0),
                ],
                "EndTimeLocal": [
                    None,
                    datetime(2024, 2, 26, 11, 5),
                    datetime(2024, 3, 31, 4, 0),
                ],
                "Error": [None, None, None],
                "RequestId": [
                    "191cc73a-fc75-4eff-9ffe-67aa5290d7f4",
                    "74d25c0b-0473-4dd9-96ff-3ca737b072a7",
                    "aec28227-f7af-4c2d-a4e6-fcb01cd570ec",
                ],
                "RefreshAttempts": None,
            }
        )
        expected["Id"] = expected["Id"].astype("Int64")
        expected["Seconds"] = expected["Seconds"].astype("Int64")

        # Act
        sut = SparkPandasDataFrame(
            json,
            schema,
            indexing_columns="Id",
            sorting_columns=["StartTimeLocal", "Id"],
            local_timezone_name="Europe/Copenhagen",
        )

        pd.options.display.max_columns = None
        pd.options.display.max_rows = None

        # Assert
        assert_frame_equal(
            expected.reset_index(drop=True), sut.get_pandas_df().reset_index(drop=True)
        )

    def test_parse_time_success(self):
        # Arrange
        data = [
            (None, None),
            (float("NaN"), None),
            ("2024-05-02", datetime(2024, 5, 2, 0, 0, 0, 0, tzinfo=None)),
            ("2024-05-03T13:35", datetime(2024, 5, 3, 13, 35, 0, 0, tzinfo=None)),
            (
                "2024-05-04T15:35:28+00:00",
                datetime(2024, 5, 4, 15, 35, 28, 0, tzinfo=None),
            ),
            (
                "2024-05-05T17:15:28.232223",
                datetime(2024, 5, 5, 17, 15, 28, 0, tzinfo=None),
            ),
            (
                "2024-05-05T21:33:04.232223+02:00",
                datetime(2024, 5, 5, 19, 33, 4, 0, tzinfo=None),
            ),
            (
                "2024-05-07T08:46:17.230735322332Z",
                datetime(2024, 5, 7, 8, 46, 17, 0, tzinfo=None),
            ),
        ]

        # Act
        for value, result in data:
            with self.subTest(value=value, result=result):
                # Act
                expected_result = SparkPandasDataFrame._parse_time(value)

                # Assert
                self.assertEqual(expected_result, result)



from types import ModuleType
from typing import List, Union

from spetlr.etl import Transformer, dataset_group
from spetlr.spark import Spark
from spetlr.sql import SqlExecutor


class SimpleSqlTransformer(Transformer):
    """This transformer takes as the input the module object
    or import path to where the sql files are that execute
    the work of this transformer.

    All dataset_input_keys will be created as spark temporary views.
    The sql must create a temporary view that has the name in dataset_output_key.
    All sql code int the module will be executed.
    """

    def __init__(
        self,
        *,
        sql_modue: Union[str, ModuleType],
        sql_file_pattern: str = "*",
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.executor = SqlExecutor(base_module=sql_modue)
        self.sql_file_pattern = sql_file_pattern

    def etl(self, inputs: dataset_group) -> dataset_group:
        # If dataset_input_keys is None, it will be set to all keys in inputs
        dataset_input_keys = self.dataset_input_keys or list(inputs.keys())

        for key in dataset_input_keys:
            inputs[key].createOrReplaceTempView(key)

        self.executor.execute_sql_file(self.sql_file_pattern)

        if self.consume_inputs:
            for key in dataset_input_keys:
                inputs.pop(key)

        # dataset_output_key is added to inputs after pop is called
        # This ensures that dataset_output_key is not popped.
        inputs[self.dataset_output_key] = Spark.get().table(self.dataset_output_key)

        return inputs



import unittest
from unittest.mock import patch

from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.deltaspec.DeltaTableSpec import DeltaTableSpec
from spetlr.deltaspec.exceptions import InvalidSpecificationError


class TestDeltaTableSpecFromSQL(unittest.TestCase):
    @patch("spetlr.configurator.sql.parse_sql.parse_single_sql_statement")
    @patch("spetlr.schema_manager.spark_schema.get_schema")
    def test_from_sql_valid(self, mock_get_schema, mock_parse_sql):
        """Test DeltaTableSpec.from_sql with a valid CREATE TABLE statement."""

        # Mock SQL parsing result
        mock_parse_sql.return_value = {
            "name": "test_db.test_table",
            "format": "delta",
            "schema": {"sql": "id INT, name STRING"},
            "path": "abfss://catalog@storage.dfs.core.windows.net/tables/test_table",
            "comment": "Test Delta Table",
            "options": {"mergeSchema": "true"},
            "partitioned_by": ["id"],
            "cluster_by": [],
            "tblproperties": {
                "delta.minReaderVersion": "2",
                "delta.minWriterVersion": "5",
            },
        }

        # Use a real StructType schema instead of a mock
        expected_schema = StructType(
            [
                StructField("id", IntegerType(), True),
                StructField("name", StringType(), True),
            ]
        )
        mock_get_schema.return_value = expected_schema

        # Input SQL
        sql = """
        CREATE TABLE test_db.test_table (
            id INT,
            name STRING
        )
        USING DELTA
        LOCATION 'abfss://catalog@storage.dfs.core.windows.net/tables/test_table'
        COMMENT 'Test Delta Table'
        OPTIONS ('mergeSchema'='true')
        PARTITIONED BY (id)
        TBLPROPERTIES ('delta.minReaderVersion'='2', 'delta.minWriterVersion'='5')
        """

        # Run method
        spec = DeltaTableSpec.from_sql(sql)

        # Assertions
        self.assertEqual(spec.name, "test_db.test_table")
        self.assertEqual(
            spec.schema, expected_schema
        )  # Now correctly compares StructType
        self.assertEqual(
            spec.location,
            "abfss://catalog@storage.dfs.core.windows.net/tables/test_table",
        )
        self.assertEqual(spec.comment, "Test Delta Table")
        self.assertEqual(spec.options, {"mergeSchema": "true"})
        self.assertEqual(spec.partitioned_by, ["id"])
        self.assertEqual(spec.cluster_by, [])

        expected_tblproperties = {
            "delta.minReaderVersion": "2",
            "delta.minWriterVersion": "5",
            "delta.columnMapping.mode": "name",
        }
        self.assertEqual(spec.tblproperties, expected_tblproperties)

    def test_from_sql_invalid_format(self):
        """Test DeltaTableSpec.from_sql with non-delta format (should raise InvalidSpecificationError)."""

        sql = """
        CREATE TABLE test_db.test_table (
            id INT,
            name STRING
        )
        USING PARQUET
        """

        with self.assertRaises(InvalidSpecificationError):
            DeltaTableSpec.from_sql(sql)

    from pyspark.sql.types import IntegerType, StructField, StructType

    def test_from_sql_missing_fields(self):
        """Test DeltaTableSpec.from_sql with missing optional fields."""

        sql = """
        CREATE TABLE test_db.test_table (
            id INT
        )
        USING DELTA
        """

        with patch(
            "spetlr.configurator.sql.parse_sql.parse_single_sql_statement"
        ) as mock_parse_sql, patch(
            "spetlr.schema_manager.spark_schema.get_schema"
        ) as mock_get_schema:
            mock_parse_sql.return_value = {
                "name": "test_db.test_table",
                "format": "delta",
                "schema": {"sql": "id INT"},
                "location": None,  # Explicitly setting location to None
                "comment": None,  # Explicitly setting comment to None
                "options": {},
                "partitioned_by": [],
                "cluster_by": [],
                "tblproperties": {},  # No tblproperties provided, so defaults should be applied
            }

            # Use a real StructType instead of a string
            expected_schema = StructType([StructField("id", IntegerType(), True)])
            mock_get_schema.return_value = expected_schema

            spec = DeltaTableSpec.from_sql(sql)

            # Assertions for missing values
            self.assertEqual(spec.name, "test_db.test_table")
            self.assertEqual(
                spec.schema, expected_schema
            )  # Now correctly compares StructType
            self.assertIsNone(spec.location)  # Should be None when not provided
            self.assertIsNone(spec.comment)  # Should be None when not provided
            self.assertEqual(spec.options, {})
            self.assertEqual(spec.partitioned_by, [])
            self.assertEqual(spec.cluster_by, [])

            # Assert that default tblproperties are set correctly
            expected_tblproperties = {
                "delta.minReaderVersion": "2",
                "delta.minWriterVersion": "5",
                "delta.columnMapping.mode": "name",
            }
            self.assertEqual(spec.tblproperties, expected_tblproperties)


if __name__ == "__main__":
    unittest.main()
